"""
    Copyright 2024 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import asyncio
import collections
import concurrent.futures
import concurrent.futures.thread
import datetime
import functools
import logging
import logging.config
import multiprocessing
import os
import socket
import typing
import uuid
from asyncio import Future, transports
from typing import Optional

import inmanta.agent.cache
import inmanta.agent.executor
import inmanta.agent.in_process_executor
import inmanta.config
import inmanta.const
import inmanta.env
import inmanta.loader
import inmanta.logging
import inmanta.protocol
import inmanta.protocol.ipc_light
import inmanta.signals
import inmanta.util
from inmanta.agent import executor
from inmanta.data.model import ResourceType
from inmanta.protocol.ipc_light import (
    FinalizingIPCClient,
    IPCMethod,
    IPCReplyFrame,
    IPCServer,
    LogReceiver,
    LogShipper,
    ReturnType,
)
from setproctitle import setproctitle

LOGGER = logging.getLogger(__name__)


class ExecutorContext:
    """The context object used by the executor to expose state to the incoming calls"""

    client: typing.Optional[inmanta.protocol.SessionClient]
    venv: typing.Optional[inmanta.env.VirtualEnv]
    name: str
    executor: typing.Optional["inmanta.agent.in_process_executor.InProcessExecutor"]

    def __init__(self, server: "ExecutorServer") -> None:
        self.server = server
        self.threadpool = concurrent.futures.thread.ThreadPoolExecutor()
        self.name = server.name

    async def stop(self) -> None:
        """Request the executor to stop"""
        if self.executor:
            self.executor.stop()
        await self.server.stop()
        if self.executor:
            # threadpool finalizer is not used, we expect threadpools to be terminated with the process
            self.executor.join([])


class ExecutorServer(IPCServer[ExecutorContext]):
    """The IPC server running on the executor

    When connected, this server will capture all logs and transport them to the remote side

    Shutdown sequence and responses

    1. Client: send stop to serverside
    2. Server: drop the connection (will drain the buffer)
    3a. Server: connection_lost is called, drop out of ioloop, exit process
    3b. Client: connection_lost calls into force_stop
    4.  Client: send term / join with timeout of grace_time / send kill / join
    5.  Client: clean up process

    Scenarios
    Server side stops: go to 3b immediately
    Client side stops: ?
    Pipe break: go to 3a and 3b
    """

    def __init__(self, name: str, logger: logging.Logger, take_over_logging: bool = True) -> None:
        """
        :param take_over_logging: when we are connected and are able to stream logs, do we remove all other log handlers?
        """
        super().__init__(name)
        self.stopping = False
        self.stopped = asyncio.Event()
        self.ctx = ExecutorContext(self)
        self.log_transport: typing.Optional[LogShipper] = None
        self.take_over_logging = take_over_logging
        self.logger = logger

    def set_status(self, status: str) -> None:
        """Update the process name to reflect the identity and status of this executor"""
        set_executor_status(self.name, status)

    def connection_made(self, transport: transports.Transport) -> None:
        super().connection_made(transport)

        # Second stage logging setup
        # Take over logger to ship to remote

        if self.take_over_logging:
            # Remove all loggers
            root_logger = logging.root
            for handler in root_logger.handlers:
                root_logger.removeHandler(handler)

        self.log_transport = LogShipper(self, asyncio.get_running_loop())
        logging.getLogger().addHandler(self.log_transport)
        self.logger.info(f"Started executor with PID: {os.getpid()}")
        self.set_status("connected")

    def _detach_log_shipper(self) -> None:
        # Once connection is lost, we want to detach asap to keep the logging clean and efficient
        if self.log_transport:
            logging.getLogger().removeHandler(self.log_transport)
            self.log_transport = None

    def get_context(self) -> ExecutorContext:
        return self.ctx

    async def stop(self) -> None:
        """Perform shutdown"""
        self._sync_stop()

    def _sync_stop(self) -> None:
        """Actual shutdown, not async"""
        if not self.stopping:
            # detach logger
            self._detach_log_shipper()
            self.logger.info("Stopping")
            self.stopping = True
            assert self.transport is not None  # Mypy
            self.transport.close()

    def connection_lost(self, exc: Exception | None) -> None:
        """We lost connection to the controler, bail out"""
        self._detach_log_shipper()
        self.logger.info("Connection lost", exc_info=exc)
        self.set_status("disconnected")
        self._sync_stop()
        self.stopped.set()


class ExecutorClient(FinalizingIPCClient[ExecutorContext], LogReceiver):
    def __init__(self, name: str):
        super().__init__(name)

        # Keeps track of when this client was active last
        self.last_used_at: datetime.datetime = datetime.datetime.now().astimezone()

    def get_idle_time(self) -> datetime.timedelta:
        return datetime.datetime.now().astimezone() - self.last_used_at

    @typing.overload
    def call(
        self, method: IPCMethod[ExecutorContext, ReturnType], has_reply: typing.Literal[True] = True
    ) -> Future[ReturnType]: ...

    @typing.overload
    def call(self, method: IPCMethod[ExecutorContext, ReturnType], has_reply: typing.Literal[False]) -> None: ...

    @typing.overload
    def call(self, method: IPCMethod[ExecutorContext, ReturnType], has_reply: bool = True) -> Future[ReturnType] | None: ...

    def call(self, method: IPCMethod[ExecutorContext, ReturnType], has_reply: bool = True) -> Future[ReturnType] | None:
        """Call a method with given arguments"""
        self.last_used_at = datetime.datetime.now().astimezone()
        response = super().call(method, has_reply)
        assert response is None or isinstance(response, Future)
        return response

    def has_outstanding_calls(self) -> bool:
        """Is this client still waiting for replies"""
        return len(self.requests) > 0

    def process_reply(self, frame: IPCReplyFrame) -> None:
        super().process_reply(frame)
        self.last_used_at = datetime.datetime.now().astimezone()


class StopCommand(inmanta.protocol.ipc_light.IPCMethod[ExecutorContext, None]):
    async def call(self, context: ExecutorContext) -> None:
        await context.stop()


class InitCommand(inmanta.protocol.ipc_light.IPCMethod[ExecutorContext, typing.Sequence[inmanta.loader.ModuleSource]]):
    """
    Initialize the executor:
    1. setup the client, using the session id of the agent
    2. activate the venv created for this executor
    3. load additional source files

    :return: module source we could not load
    """

    def __init__(
        self,
        venv_path: str,
        storage_folder: str,
        session_gid: uuid.UUID,
        sources: list[inmanta.loader.ModuleSource],
        environment: uuid.UUID,
        uri: str,
    ):
        self.venv_path = venv_path
        self.storage_folder = storage_folder
        self.gid = session_gid
        self.sources = sources
        self.environment = environment
        self.uri = uri

    async def call(self, context: ExecutorContext) -> typing.Sequence[inmanta.loader.ModuleSource]:
        loop = asyncio.get_running_loop()
        parent_logger = logging.getLogger("agent.executor")
        logger = parent_logger.getChild(context.name)

        # setup client
        context.client = inmanta.protocol.SessionClient("agent", self.gid)

        # Setup agent instance
        context.executor = inmanta.agent.in_process_executor.InProcessExecutor(
            agent_name=context.name,
            agent_uri=self.uri,
            environment=self.environment,
            client=context.client,
            eventloop=loop,
            parent_logger=parent_logger,
        )

        # activate venv
        context.venv = inmanta.env.VirtualEnv(self.venv_path)
        context.venv.use_virtual_env()

        # Download and load code
        loader = inmanta.loader.CodeLoader(self.storage_folder)

        sync_client = inmanta.protocol.SyncClient(client=context.client, ioloop=loop)
        sources = [s.with_client(sync_client) for s in self.sources]

        failed: list[inmanta.loader.ModuleSource] = []
        in_place: list[inmanta.loader.ModuleSource] = []
        # First put all files on disk
        for module_source in sources:
            try:
                await loop.run_in_executor(context.threadpool, functools.partial(loader.install_source, module_source))
                in_place.append(module_source)
            except Exception:
                logger.info("Failed to load sources: %s", module_source, exc_info=True)
                failed.append(module_source)

        # then try to import them
        for module_source in in_place:
            try:
                await loop.run_in_executor(
                    context.threadpool,
                    functools.partial(loader._load_module, module_source.name, module_source.hash_value, require_reload=False),
                )
            except Exception:
                logger.info("Failed to load sources: %s", module_source, exc_info=True)
                failed.append(module_source)

        return failed


class OpenVersionCommand(inmanta.protocol.ipc_light.IPCMethod[ExecutorContext, None]):

    def __init__(self, version: int) -> None:
        self.version = version

    async def call(self, context: ExecutorContext) -> None:
        assert context.executor is not None
        await context.executor.open_version(self.version)


class CloseVersionCommand(inmanta.protocol.ipc_light.IPCMethod[ExecutorContext, None]):

    def __init__(self, version: int) -> None:
        self.version = version

    async def call(self, context: ExecutorContext) -> None:
        assert context.executor is not None
        await context.executor.close_version(self.version)


class DryRunCommand(inmanta.protocol.ipc_light.IPCMethod[ExecutorContext, None]):

    def __init__(
        self,
        resources: typing.Sequence["inmanta.agent.executor.ResourceDetails"],
        dry_run_id: uuid.UUID,
    ) -> None:
        self.resources = resources
        self.dry_run_id = dry_run_id

    async def call(self, context: ExecutorContext) -> None:
        assert context.executor is not None
        await context.executor.dry_run(self.resources, self.dry_run_id)


class ExecuteCommand(inmanta.protocol.ipc_light.IPCMethod[ExecutorContext, None]):

    def __init__(
        self,
        gid: uuid.UUID,
        resource_details: "inmanta.agent.executor.ResourceDetails",
        reason: str,
    ) -> None:
        self.gid = gid
        self.resource_details = resource_details
        self.reason = reason

    async def call(self, context: ExecutorContext) -> None:
        assert context.executor is not None
        await context.executor.execute(self.gid, self.resource_details, self.reason)


class FactsCommand(inmanta.protocol.ipc_light.IPCMethod[ExecutorContext, inmanta.types.Apireturn]):

    def __init__(self, resource: "inmanta.agent.executor.ResourceDetails") -> None:
        self.resource = resource

    async def call(self, context: ExecutorContext) -> inmanta.types.Apireturn:
        assert context.executor is not None
        return await context.executor.get_facts(self.resource)


def set_executor_status(name: str, status: str) -> None:
    """Update the process name to reflect the identity and status of the executor"""
    # Lives outside the ExecutorServer class, so we can set status early in the boot process
    setproctitle(f"inmanta: executor {name} - {status}")


def mp_worker_entrypoint(
    socket: socket.socket,
    name: str,
    log_level: int,
    cli_log: bool,
    config: typing.Mapping[str, typing.Mapping[str, typing.Any]],
) -> None:
    """Entry point for child processes"""

    set_executor_status(name, "connecting")
    # Set up logging stage 1
    # Basic config, starts on std.out
    config_builder = inmanta.logging.LoggingConfigBuilder()
    logger_config: inmanta.logging.FullLoggingConfig = config_builder.get_bootstrap_logging_config(python_log_level=log_level)
    logger_config.apply_config()
    logging.captureWarnings(True)

    # Set up our own logger
    logger = logging.getLogger(f"agent.executor.{name}")

    # Load config
    inmanta.config.Config.load_config_from_dict(config)

    async def serve() -> None:
        loop = asyncio.get_running_loop()
        # Start serving
        # also performs setup of log shipper
        # this is part of stage 2 logging setup
        transport, protocol = await loop.connect_accepted_socket(
            functools.partial(ExecutorServer, name, logger, not cli_log), socket
        )
        inmanta.signals.setup_signal_handlers(protocol.stop)
        await protocol.stopped.wait()

    # Async init
    asyncio.run(serve())
    logger.info(f"Stopped with PID: {os.getpid()}")
    exit(0)


class MPExecutor(executor.Executor):
    """A Single Child Executor"""

    def __init__(
        self,
        owner: "MPManager",
        process: multiprocessing.Process,
        connection: ExecutorClient,
        executor_id: executor.ExecutorId,
        venv: executor.ExecutorVirtualEnvironment,
    ):
        self.process = process
        self.connection = connection
        self.connection.finalizers.append(self.force_stop)
        self.closing = False
        self.closed = False
        self.owner = owner
        # Pure for debugging purpose
        self.executor_id = executor_id
        self.executor_virtual_env = venv

        # Set by init and parent class that const
        self.failed_resource_sources: typing.Sequence[inmanta.loader.ModuleSource] = list()
        self.failed_resource_types: set[ResourceType] = set()

    async def stop(self) -> None:
        """Stop by shutdown"""
        self.closing = True
        try:
            self.connection.call(StopCommand(), False)
        except inmanta.protocol.ipc_light.ConnectionLost:
            # Already gone
            pass

    def can_be_cleaned_up(self, retention_time: int) -> bool:
        if self.connection.has_outstanding_calls():
            return False

        return self.connection.get_idle_time() > datetime.timedelta(seconds=retention_time)

    def last_used(self) -> datetime.datetime:
        return self.connection.last_used_at

    async def force_stop(self, grace_time: float = inmanta.const.SHUTDOWN_GRACE_HARD) -> None:
        """Stop by process close"""
        self.closing = True
        await asyncio.get_running_loop().run_in_executor(
            self.owner.thread_pool, functools.partial(self._force_stop, grace_time)
        )

    def _force_stop(self, grace_time: float) -> None:
        if self.closed:
            return
        try:
            self.process.terminate()
            self.process.join(grace_time)
            self.process.kill()
            self.process.join()
        except ValueError:
            # Process already gone:
            pass
        self._set_closed()

    def _set_closed(self) -> None:
        # this code can be raced from the join call and the disconnect handler
        # relying on the GIL to keep us safe
        if not self.closed:
            self.closing = True
            self.closed = True
            self.process.close()
            self.owner._child_closed(self)

    async def join(self, timeout: float) -> None:
        if self.closed:
            return
        try:
            await asyncio.get_running_loop().run_in_executor(
                self.owner.thread_pool, functools.partial(self.process.join, timeout)
            )
            if self.process.exitcode is None:
                LOGGER.warning(
                    "Executor for agent %s with id %s didn't stop after timeout of %d seconds. Killing it.",
                    self.executor_id.agent_name,
                    self.executor_id.identity(),
                    timeout,
                )
                self.process.kill()
                await asyncio.get_running_loop().run_in_executor(
                    self.owner.thread_pool, functools.partial(self.process.join, timeout)
                )
            self._set_closed()
        except ValueError as e:
            if "process object is closed" in str(e):
                # process already closed
                # raises a value error, so we also check the message
                pass
            else:
                raise

    async def close_version(self, version: int) -> None:
        await self.connection.call(CloseVersionCommand(version))

    async def open_version(self, version: int) -> None:
        await self.connection.call(OpenVersionCommand(version))

    async def dry_run(
        self,
        resources: typing.Sequence["inmanta.agent.executor.ResourceDetails"],
        dry_run_id: uuid.UUID,
    ) -> None:
        await self.connection.call(DryRunCommand(resources, dry_run_id))

    async def execute(
        self,
        gid: uuid.UUID,
        resource_details: "inmanta.agent.executor.ResourceDetails",
        reason: str,
    ) -> None:
        await self.connection.call(ExecuteCommand(gid, resource_details, reason))

    async def get_facts(self, resource: "inmanta.agent.executor.ResourceDetails") -> inmanta.types.Apireturn:
        return await self.connection.call(FactsCommand(resource))


class MPManager(executor.ExecutorManager[MPExecutor]):
    """
    This is the executor that provides the new behavior (ISO8+),
    where the agent forks executors in specific venvs to prevent code reloading.
    """

    def __init__(
        self,
        thread_pool: concurrent.futures.thread.ThreadPoolExecutor,
        environment_manager: executor.VirtualEnvironmentManager,
        session_gid: uuid.UUID,
        environment: uuid.UUID,
        log_folder: str,
        storage_folder: str,
        log_level: int = logging.INFO,
        cli_log: bool = False,
    ) -> None:
        """
        :param thread_pool:  threadpool to perform work on
        :param environment_manager: The VirtualEnvironmentManager responsible for managing the virtual environments
        :param session_gid: agent session id, used to connect to the server, the agent should keep this alive
        :param environment: the inmanta environment we are deploying for
        :param log_folder: folder to place log files for the executors
        :param storage_folder: folder to place code files
        :param log_level: log level for the executors
        :param cli_log: do we also want to echo the log to std_err

        """
        self.init_once()
        self.environment = environment
        self.thread_pool = thread_pool
        self.environment_manager = environment_manager
        self.children: list[MPExecutor] = []
        self.log_folder = log_folder
        self.storage_folder = storage_folder
        os.makedirs(self.log_folder, exist_ok=True)
        os.makedirs(self.storage_folder, exist_ok=True)
        self.log_level = log_level
        self.cli_log = cli_log
        self.session_gid = session_gid

        # These maps are cleaned by the close callbacks
        # This means it can contain closing entries
        self.executor_map: dict[executor.ExecutorId, MPExecutor] = {}
        self.agent_map: dict[str, set[executor.ExecutorId]] = collections.defaultdict(set)

        self._locks: inmanta.util.NamedLock = inmanta.util.NamedLock()

        self.executor_retention_time = inmanta.agent.config.agent_executor_retention_time.get()
        self.max_executors_per_agent = inmanta.agent.config.agent_executor_cap.get()

        # We keep a reference to the periodic cleanup task to prevent it
        # from disappearing mid-execution https://docs.python.org/3.11/library/asyncio-task.html#creating-tasks
        self.cleanup_job: Optional[asyncio.Task[None]] = None

    def __add_executor(self, theid: executor.ExecutorId, the_executor: MPExecutor) -> None:
        self.executor_map[theid] = the_executor
        self.agent_map[theid.agent_name].add(theid)

    def __remove_executor(self, the_executor: MPExecutor) -> None:
        theid = the_executor.executor_id
        registered_for_id = self.executor_map.get(theid)
        if registered_for_id is None:
            # Not found
            return
        if registered_for_id != the_executor:
            # We have a stale instance that refused to close before
            # it was replaced and has now gone down
            return
        self.executor_map.pop(theid)
        self.agent_map[theid.agent_name].discard(theid)

    @classmethod
    def init_once(cls) -> None:
        try:
            multiprocessing.set_start_method("forkserver")
            # Load common modules
            # Including this one
            multiprocessing.set_forkserver_preload(["inmanta.config", __name__, "inmanta.agent._set_fork_server_process_name"])
        except RuntimeError:
            # already set
            pass

    async def get_executor(
        self, agent_name: str, agent_uri: str, code: typing.Collection[executor.ResourceInstallSpec]
    ) -> MPExecutor:
        """
        Retrieves an Executor for a given agent with the relevant handler code loaded in its venv.
        If an Executor does not exist for the given configuration, a new one is created.

        :param agent_name: The name of the agent for which an Executor is being retrieved or created.
        :param agent_uri: The name of the host on which the agent is running.
        :param code: Collection of ResourceInstallSpec defining the configuration for the Executor i.e.
            which resource types it can act on and all necessary information to install the relevant
            handler code in its venv.
        :return: An Executor instance
        """
        blueprint = executor.ExecutorBlueprint.from_specs(code)
        executor_id = executor.ExecutorId(agent_name, agent_uri, blueprint)
        if executor_id in self.executor_map:
            it = self.executor_map[executor_id]
            if not it.closing:
                LOGGER.debug("Found existing executor for agent %s with id %s", agent_name, executor_id.identity())
                return it
        # Acquire a lock based on the executor's identity (agent name, agent uri and blueprint hash)
        async with self._locks.get(executor_id.identity()):
            if executor_id in self.executor_map:
                it = self.executor_map[executor_id]
                if not it.closing:
                    LOGGER.debug("Found existing executor for agent %s with id %s", agent_name, executor_id.identity())
                    return it
                else:
                    LOGGER.debug(
                        "Found stale executor for agent %s with id %s, waiting for close", agent_name, executor_id.identity()
                    )
                    await it.join(2.0)
            n_executors_for_agent = len(self.agent_map[executor_id.agent_name])
            if n_executors_for_agent >= self.max_executors_per_agent:
                # Close oldest executor:
                executor_ids = self.agent_map[executor_id.agent_name]
                oldest_executor = min([self.executor_map[id] for id in executor_ids], key=lambda e: e.connection.last_used_at)
                LOGGER.debug(
                    f"Reached executor cap for agent {executor_id.agent_name}. Stopping oldest executor "
                    f"{oldest_executor.executor_id.identity()} to make room for a new one."
                )

                await oldest_executor.stop()

            my_executor = await self.create_executor(executor_id)
            self.__add_executor(executor_id, my_executor)
            if my_executor.failed_resource_sources:
                # If some code loading failed, resolve here
                # reverse index
                type_for_spec: dict[inmanta.loader.ModuleSource, list[ResourceType]] = collections.defaultdict(list)
                for spec in code:
                    for source in spec.blueprint.sources:
                        type_for_spec[source].append(spec.resource_type)
                # resolve
                for source in my_executor.failed_resource_sources:
                    for rtype in type_for_spec.get(source, []):
                        my_executor.failed_resource_types.add(rtype)

            # TODO: recovery. If loading failed, we currently never rebuild
            # https://github.com/inmanta/inmanta-core/issues/7281
            return my_executor

    async def create_executor(self, executor_id: executor.ExecutorId) -> MPExecutor:
        LOGGER.info("Creating executor for agent %s with id %s", executor_id.agent_name, executor_id.identity())
        env_blueprint = executor_id.blueprint.to_env_blueprint()
        venv = await self.environment_manager.get_environment(env_blueprint, self.thread_pool)
        executor = await self.make_child_and_connect(executor_id, venv)
        LOGGER.debug(
            "Child forked (pid: %s) for executor for agent %s with id %s",
            executor.process.pid,
            executor_id.agent_name,
            executor_id.identity(),
        )
        storage_for_blueprint = os.path.join(self.storage_folder, "code", executor_id.blueprint.blueprint_hash())
        os.makedirs(storage_for_blueprint, exist_ok=True)
        failed_types = await executor.connection.call(
            InitCommand(
                venv.env_path,
                storage_for_blueprint,
                self.session_gid,
                [x.for_transport() for x in executor_id.blueprint.sources],
                self.environment,
                executor_id.agent_uri,
            )
        )
        LOGGER.debug(
            "Child initialized (pid: %s) for executor for agent %s with id %s",
            executor.process.pid,
            executor_id.agent_name,
            executor_id.identity(),
        )
        executor.failed_resource_sources = failed_types
        return executor

    async def make_child_and_connect(
        self, executor_id: executor.ExecutorId, venv: executor.ExecutorVirtualEnvironment
    ) -> MPExecutor:
        """Async code to make a child process and share a socket with it"""
        loop = asyncio.get_running_loop()
        name = executor_id.agent_name

        # Start child
        process, parent_conn = await loop.run_in_executor(
            self.thread_pool, functools.partial(self._make_child, name, self.log_level, self.cli_log)
        )
        # Hook up the connection
        transport, protocol = await loop.connect_accepted_socket(
            functools.partial(ExecutorClient, f"executor.{name}"), parent_conn
        )

        child_handle = MPExecutor(self, process, protocol, executor_id, venv)
        self.children.append(child_handle)
        return child_handle

    def _child_closed(self, child_handle: MPExecutor) -> None:
        """Internal, for child to remove itself once stopped"""
        try:
            self.children.remove(child_handle)
            self.__remove_executor(child_handle)
        except ValueError:
            # already gone
            pass

    def _make_child(self, name: str, log_level: int, cli_log: bool) -> tuple[multiprocessing.Process, socket.socket]:
        """Sync code to make a child process and share a socket with it"""
        parent_conn, child_conn = socket.socketpair()
        # Fork an ExecutorServer
        p = multiprocessing.Process(
            target=mp_worker_entrypoint,
            args=(child_conn, name, log_level, cli_log, inmanta.config.Config.config_as_dict()),
            name=f"agent.executor.{name}",
        )
        p.start()
        child_conn.close()

        return p, parent_conn

    async def start(self) -> None:
        self.running = True
        self.cleanup_job = asyncio.create_task(self.cleanup_inactive_executors())

    async def stop(self) -> None:
        self.running = False
        await asyncio.gather(*(child.stop() for child in self.children))

    async def force_stop(self, grace_time: float) -> None:
        await asyncio.gather(*(child.force_stop(grace_time) for child in self.children))

    async def join(self, thread_pool_finalizer: list[concurrent.futures.ThreadPoolExecutor], timeout: float) -> None:
        thread_pool_finalizer.append(self.thread_pool)
        await asyncio.gather(*(child.join(timeout) for child in self.children))
        if self.cleanup_job:
            await self.cleanup_job

    async def stop_for_agent(self, agent_name: str) -> list[MPExecutor]:
        children_ids = self.agent_map[agent_name]
        children = [self.executor_map[child_id] for child_id in children_ids]
        await asyncio.gather(*(child.stop() for child in children))
        return children

    async def cleanup_inactive_executors(self) -> None:
        """
        This task periodically cleans up idle executors
        """
        while self.running:
            cleanup_start = datetime.datetime.now().astimezone()

            reschedule_interval: float = self.executor_retention_time
            for _executor in self.executor_map.values():
                if _executor.can_be_cleaned_up(self.executor_retention_time):
                    async with self._locks.get(_executor.executor_id.identity()):
                        # Check that the executor can still be cleaned up by the time we have acquired the lock
                        if _executor.can_be_cleaned_up(self.executor_retention_time):
                            try:
                                LOGGER.debug(
                                    "Stopping executor %s because it "
                                    "was inactive for %d s, which is longer then the retention time of %d s.",
                                    _executor.executor_id.identity(),
                                    _executor.connection.get_idle_time().total_seconds(),
                                    self.executor_retention_time,
                                )
                                await _executor.stop()
                            except Exception:
                                LOGGER.debug(
                                    "Unexpected error during executor %s cleanup:",
                                    _executor.executor_id.identity(),
                                    exc_info=True,
                                )
                else:
                    reschedule_interval = min(
                        reschedule_interval,
                        (
                            datetime.timedelta(seconds=self.executor_retention_time)
                            - (cleanup_start - _executor.connection.last_used_at)
                        ).total_seconds(),
                    )

            cleanup_end = datetime.datetime.now().astimezone()

            await asyncio.sleep(max(0.0, reschedule_interval - (cleanup_end - cleanup_start).total_seconds()))



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import asyncio
import re

import pytest
import tornado
from pyformance import gauge, global_registry, timer
from tornado.httpserver import HTTPServer
from tornado.web import url

from inmanta.reporter import AsyncReporter, InfluxReporter
from inmanta.server.services.metricservice import CPUMicroBenchMark


class QueryMockHandler(tornado.web.RequestHandler):
    def initialize(self, parent):
        self.parent = parent

    def get(self, *args, **kwargs):
        self.parent.querycount += 1
        try:
            assert self.request.query_arguments["q"] == [b"CREATE DATABASE metrics"]
        except Exception as e:
            # carry over  failures
            self.parent.failure = e


influxlineprotocol = re.compile(r"\w+(,\w+=\w+)* (\w+=[\d.e+-]*)(,\w+=[\d.e+-]*)* \d+")


class WriteMockHandler(tornado.web.RequestHandler):
    def initialize(self, parent):
        self.parent = parent

    def post(self, *args, **kwargs):
        self.parent.writecount += 1
        try:
            for line in self.request.body.decode().split("\n"):
                print(line)
                assert influxlineprotocol.match(line)
                self.parent.lines.append(line)
        except Exception as e:
            # carry over  failures
            self.parent.failure = e


class InfluxdbMock:
    def __init__(self, socket):
        self.querycount = 0
        self.writecount = 0
        self.failure = None
        self.lines = []

        self.app = tornado.web.Application(
            [
                url(r"/query", QueryMockHandler, kwargs={"parent": self}),
                url(r"/write", WriteMockHandler, kwargs={"parent": self}),
            ]
        )

        self.server = HTTPServer(self.app)
        self.server.add_sockets([socket])
        _addr, port = socket.getsockname()
        self.port = port


class MockReporter(AsyncReporter):
    def __init__(self, interval):
        super().__init__(None, interval)
        self.waiter = asyncio.locks.Semaphore(0)
        self.in_count = 0
        self.count = 0

    async def report_now(self, registry=None, timestamp=None) -> None:
        self.in_count += 1
        await self.waiter.acquire()
        self.count += 1


@pytest.fixture
def influxdb(event_loop, free_socket):
    ifl = InfluxdbMock(free_socket())
    yield ifl
    ifl.server.stop()


async def test_influxdb(influxdb):
    rep = InfluxReporter(port=influxdb.port, tags={"mark": "X"}, autocreate_database=True)
    with timer("test").time():
        pass

    with timer("test2").time():
        pass

    cpu_micro = CPUMicroBenchMark()
    gauge("CPU", cpu_micro)
    await rep.report_now()

    assert influxdb.querycount == 1
    assert influxdb.writecount == 1

    # cpu micro has none-empty cache
    assert cpu_micro.last_value is not None

    if influxdb.failure:
        raise influxdb.failure

    for line in influxdb.lines:
        assert "mark=X" in line


async def test_timing():
    # Attempt to deploy every 0.01 seconds,
    # Deploy hangs on a semaphore, so test case can control progress
    mr = MockReporter(0.01)
    mr.start()

    # no deploy done
    assert mr.count == 0
    # wait for deploy
    await asyncio.sleep(0.01)
    # one is waiting
    assert mr.in_count == 1
    # release
    mr.waiter.release()
    # wait 0 to allow reporter to make progress
    await asyncio.sleep(0.0)
    # one deploy done
    assert mr.count == 1

    # allow to run free to test timing
    for i in range(5):
        mr.waiter.release()

    # wait 0 to allow reporter to make progress
    await asyncio.sleep(0)
    # how many are we now?
    base = mr.count
    # should be sufficiently far of the lock
    # otherwise test always succeeds
    assert base < 4
    # wait for one more
    await asyncio.sleep(0.01)
    assert mr.count == base + 1
    mr.stop()


async def test_available_metrics(server):
    metrics = global_registry().dump_metrics()

    types: dict[str, list[type]] = {}
    # Ensure type consistency
    for metric in metrics.values():
        for key, value in metric.items():
            if key in types:
                # don't use isinstance, because bool is int in python, but not for influxdb
                assert type(value) in types[key], f"inconsistent types for {key}: {type(value)} and {types[key]}"
            else:
                base_type = type(value)
                # https://docs.influxdata.com/influxdb/v1.8/write_protocols/line_protocol_tutorial/
                # Floats - by default, InfluxDB assumes all numerical field values are floats.
                if base_type == int or base_type == float:
                    base_types = [int, float]
                else:
                    base_types = [base_type]
                types[key] = base_types

    assert metrics["db.connected"]["value"]
    assert "db.max_pool" in metrics
    assert "db.open_connections" in metrics
    assert "db.free_connections" in metrics
    assert "self.spec.cpu" in metrics

    # ensure it doesn't crash when the server is down
    await server.stop()
    metrics = global_registry().dump_metrics()

    assert metrics["db.max_pool"]["value"] == 0
    assert not metrics["db.connected"]["value"]


async def test_safeness_if_server_down():
    # ensure it works is there is no server
    global_registry().dump_metrics()



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import argparse
import logging
import os
import socket
import subprocess
import sys
import time
from typing import Optional

from inmanta import config, const, module, postgresproc, protocol
from inmanta.config import Config
from inmanta.protocol import Result
from inmanta.types import JsonType
from inmanta.util import get_free_tcp_port

LOGGER = logging.getLogger(__name__)
MAX_TRIES = 25


cfg_prj = config.Option("deploy", "project", "deploy", "The project name to use in the deploy", config.is_str_opt)
cfg_env = config.Option("deploy", "environment", "deploy", "The environment name to use in the deploy", config.is_str_opt)


class FinishedException(Exception):
    """
    This exception is raised when the deploy is ready
    """


class Deploy:
    _data_path: str
    _project_path: str
    _server_proc: subprocess.Popen
    _postgresproc: postgresproc.PostgresProc
    _client: protocol.SyncClient
    _environment_id: str

    def __init__(self, options: argparse.Namespace, postgresport: int = 0) -> None:
        self._postgresport = postgresport
        self._server_port = 0
        self._options = options

        loud_logger = logging.getLogger("inmanta.protocol")
        loud_logger.propagate = False

        loud_logger = logging.getLogger("tornado")
        loud_logger.propagate = False

    def _check_result(self, result: protocol.Result, fatal: bool = True) -> protocol.Result:
        """Check the result of a call protocol call. If the result is not 200, issue an error"""
        if result.code != 200 and not result.result:
            msg = f"Server request failed with code {result.code} and result {result.result}"
            LOGGER.error(msg)

            if fatal:
                raise Exception(msg)

        return result

    def _ensure_dir(self, path: str) -> None:
        if not os.path.exists(path):
            LOGGER.debug("Creating directory %s", path)
            os.mkdir(path)

    def setup_server(self) -> bool:
        state_dir = os.path.join(self._data_path, "state")
        self._ensure_dir(state_dir)

        log_dir = os.path.join(self._data_path, "logs")
        self._ensure_dir(log_dir)

        self._server_port = int(get_free_tcp_port())
        assert self._server_port != self._postgresport

        config.Config.set("client_rest_transport", "port", str(self._server_port))

        vars_in_configfile = {
            "state_dir": state_dir,
            "log_dir": log_dir,
            "server_port": self._server_port,
            "postgres_port": self._postgresport,
        }
        config_file = (
            """
[database]
name=postgres
port=%(postgres_port)s
host=localhost
username=postgres

[config]
state-dir=%(state_dir)s
log-dir=%(log_dir)s

[server]
bind-port=%(server_port)s
bind-address=127.0.0.1

[agent_rest_transport]
port=%(server_port)s
host=localhost

[compiler_rest_transport]
port=%(server_port)s
host=localhost

[client_rest_transport]
port=%(server_port)s
host=localhost

[cmdline_rest_transport]
port=%(server_port)s
host=localhost
"""
            % vars_in_configfile
        )

        server_config = os.path.join(self._data_path, "server.cfg")
        with open(server_config, "w+", encoding="utf-8") as fd:
            fd.write(config_file)

        log_file = os.path.join(log_dir, "inmanta.log")
        args = [
            sys.executable,
            "-m",
            "inmanta.app",
            "--log-file-level",
            "DEBUG",
            "-c",
            server_config,
            "--config-dir",
            Config._config_dir if Config._config_dir is not None else "",
            "--log-file",
            log_file,
            "server",
        ]

        self._server_proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        LOGGER.debug("Started server on port %d", self._server_port)

        while self._server_proc.poll() is None:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                try:
                    s.connect(("localhost", int(self._server_port)))
                    return True
                except OSError:
                    time.sleep(0.25)
            finally:
                s.close()

        return False

    def setup_postgresql(self) -> bool:
        # start a local postgresql server on a random port
        if self._postgresport == 0:
            self._postgresport = int(get_free_tcp_port())
            postgres_dir = os.path.join(self._data_path, "postgres")
            self._postgresproc = postgresproc.PostgresProc(port=self._postgresport, db_path=postgres_dir)
            LOGGER.debug("Starting postgresql on port %d", self._postgresport)
            if not self._postgresproc.start():
                LOGGER.error(
                    "Unable to start postgresql instance on port %d and data directory %s", self._postgresport, postgres_dir
                )
                return False

        return True

    def _create_project(self, project_name: str) -> Optional[str]:
        LOGGER.debug("Creating project %s", project_name)
        result = self._client.create_project(project_name)
        if result.code != 200 or not result.result:
            LOGGER.error("Unable to create project %s", project_name)
            return None

        return str(result.result["project"]["id"])

    def _create_environment(self, project_id: str, environment_name: str) -> Optional[str]:
        LOGGER.debug("Creating environment %s in project %s", environment_name, project_id)
        result = self._client.create_environment(project_id=project_id, name=environment_name)
        if result.code != 200 or not result.result:
            LOGGER.error("Unable to create environment %s", environment_name)
            return None

        env_id: str = result.result["environment"]["id"]

        self._check_result(self._client.set_setting(env_id, "autostart_agent_deploy_splay_time", 0))
        self._check_result(self._client.set_setting(env_id, "autostart_agent_deploy_interval", 0))
        self._check_result(self._client.set_setting(env_id, "autostart_agent_repair_splay_time", 0))
        self._check_result(self._client.set_setting(env_id, "autostart_agent_repair_interval", 600))

        return env_id

    def _latest_version_instance(self, environment_id: str) -> Optional[JsonType]:
        result: Result = self._client.list_versions(tid=environment_id)
        if result.code != 200:
            LOGGER.error("Unable to get all version of environment %s", environment_id)
            return None

        if result.result and "versions" in result.result and len(result.result["versions"]) > 0:
            version = result.result["versions"][0]
            assert isinstance(version, dict)  # mypy
            return version

        return None

    def setup_project(self) -> bool:
        """
        Set up the configured project and environment on the embedded server
        """
        self._client = protocol.SyncClient("client")

        # get config
        project_name = cfg_prj.get()
        if project_name is None:
            LOGGER.error("The name of the project should be configured for an all-in-one deploy")
            return False

        environment_name = cfg_env.get()
        if environment_name is None:
            LOGGER.error("The name of the environment in the project should be configured for an all-in-one deploy")
            return False

        # wait and check to see if the server is up
        tries = 0
        while tries < MAX_TRIES:
            try:
                self._client.list_projects()
                break
            except Exception:
                tries += 1

        # get project id
        projects = self._client.list_projects()
        if projects.code != 200 or not projects.result:
            LOGGER.error("Unable to retrieve project listing from the server")
            return False

        project_id = None
        for project in projects.result["projects"]:
            if project_name == project["name"]:
                project_id = project["id"]
                break

        if project_id is None:
            project_id = self._create_project(project_name)
            if not project_id:
                return False

        # get or create the environment
        environments = self._client.list_environments()
        if environments.code != 200 or not environments.result:
            LOGGER.error("Unable to retrieve environments from server")
            return False

        env_id = None
        for env in environments.result["environments"]:
            if project_id == env["project"] and environment_name == env["name"]:
                env_id = env["id"]
                break

        if env_id is None:
            env_id = self._create_environment(project_id, environment_name)
            if not env_id:
                return False

        self._environment_id = env_id

        # link the project into the server environment
        server_env = os.path.join(self._data_path, "state", "server", "environments", self._environment_id)
        full_path = os.path.abspath(self._project_path)
        if not os.path.islink(server_env) or os.readlink(server_env) != full_path:
            if os.path.exists(server_env):
                os.unlink(server_env)
            os.symlink(src=full_path, dst=server_env)

        return True

    def setup(self) -> bool:
        """
        Run inmanta locally
        """
        # create local storage
        project = module.Project.get()
        self._project_path = project.project_path
        self._data_path = os.path.join(project.project_path, "data", "deploy")
        LOGGER.debug("Storing state data in %s", self._data_path)
        self._ensure_dir(os.path.join(project.project_path, "data"))
        self._ensure_dir(self._data_path)

        if not self.setup_postgresql():
            return False

        if not self.setup_server():
            return False

        if not self.setup_project():
            LOGGER.error("Failed to setup project")
            return False

        return True

    def export(self, main_file: str) -> bool:
        """
        Export a version to the embedded server
        """
        inmanta_path = [sys.executable, "-m", "inmanta.app"]

        cmd = inmanta_path + [
            "--config-dir",
            Config._config_dir if Config._config_dir is not None else "",
            "-vvv",
            "export",
            "-e",
            str(self._environment_id),
            "--server_address",
            "localhost",
            "--server_port",
            str(self._server_port),
            "-f",
            main_file,
        ]

        sub_process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        log_out, log_err = sub_process.communicate()

        if sub_process.returncode > 0:
            print("An error occurred while compiling the model:")
            if len(log_out) > 0:
                print(log_out.decode())
            if len(log_err) > 0:
                print(log_err.decode())

            return False

        LOGGER.info("Export of model complete")
        return True

    def deploy(self, dry_run: bool, report: bool = True) -> None:
        version_instance = self._latest_version_instance(self._environment_id)

        if version_instance is None:
            return

        version: int = version_instance["version"]
        LOGGER.info("Latest version for created environment is %s", version)

        # release the version!
        if not dry_run:
            if not version_instance["released"]:
                self._check_result(
                    self._client.release_version(
                        tid=self._environment_id,
                        id=version,
                        push=True,
                        agent_trigger_method=const.AgentTriggerMethod.push_full_deploy,
                    )
                )
            else:
                self._check_result(
                    self._client.deploy(
                        tid=self._environment_id,
                        agent_trigger_method=const.AgentTriggerMethod.push_full_deploy,
                    )
                )
            if report:
                self.progress_deploy_report(version)

        else:
            result = self._check_result(self._client.dryrun_request(tid=self._environment_id, id=version))
            dryrun_id = result.result["dryrun"]["id"]
            if report:
                self.progress_dryrun_report(dryrun_id)

    def _get_deploy_stats(self, version: int) -> tuple[int, int, dict[str, str]]:
        version_result = self._client.get_version(tid=self._environment_id, id=version)
        if version_result.code != 200 or not version_result.result:
            LOGGER.error("Unable to get version %d of environment %s", version, self._environment_id)
            return (0, 0, {})

        total = 0
        deployed = 0
        ready = {}

        for res in version_result.result["resources"]:
            total += 1
            if res["status"] not in [x.name for x in const.TRANSIENT_STATES]:
                deployed += 1
                ready[res["id"]] = res["status"]

        return total, deployed, ready

    def progress_deploy_report(self, version: int) -> None:
        print("Starting deploy")
        current_ready: set[str] = set()
        total = 0
        deployed = -1
        while total > deployed:
            total, deployed, ready = self._get_deploy_stats(version)

            ready_keys = set(ready.keys())
            new = ready_keys - current_ready
            current_ready = ready_keys

            # if we already printed progress, move cursor one line up
            if deployed >= 0:
                sys.stdout.write("\033[1A")
                sys.stdout.flush()

            for res in new:
                print(f"{res} - {ready[res]}")

            print("[%d / %d]" % (deployed, total))
            time.sleep(1)

        print("Deploy ready")

    def _get_dryrun_status(self, dryrun_id: str) -> tuple[int, int, JsonType]:
        result = self._client.dryrun_report(self._environment_id, dryrun_id)

        if result.code != 200 or not result.result:
            raise Exception("Unable to get dryrun report")

        data = result.result["dryrun"]
        return data["total"], data["todo"], data["resources"]

    def progress_dryrun_report(self, dryrun_id: str) -> None:
        print("Starting dryrun")

        current_ready: set[str] = set()
        todo = 1
        while todo > 0:
            # if we already printed progress, move cursor one line up
            if len(current_ready) > 0:
                sys.stdout.write("\033[1A")
                sys.stdout.flush()

            total, todo, ready = self._get_dryrun_status(dryrun_id)

            ready_keys: set[str] = set(ready.keys())
            new = ready_keys - current_ready
            current_ready = ready_keys

            for res in new:
                changes: dict[str, tuple[str, str]] = ready[res]["changes"]
                if len(changes) == 0:
                    print("%s - no changes" % res)
                else:
                    print("%s:" % res)
                    for field, values in changes.items():
                        if field == "hash":
                            diff_result = self._client.diff(a=values[0], b=values[1])
                            if diff_result.code == 200 and diff_result.result:
                                print("  - content:")
                                diff_value = diff_result.result
                                print("    " + "    ".join(diff_value["diff"]))
                        else:
                            print("  - %s:" % field)
                            print("    from: %s" % values[0])
                            print("    to:   %s" % values[1])

                        print("")

            print("[%d / %d]" % (total - todo, total))
            time.sleep(1)

        raise FinishedException()

    def run(self) -> None:
        self.export(main_file=self._options.main_file)
        self.deploy(dry_run=self._options.dryrun)

    def stop(self) -> None:
        loud_logger = logging.getLogger("inmanta.protocol")
        loud_logger.propagate = True

        loud_logger = logging.getLogger("tornado")
        loud_logger.propagate = True

        if hasattr(self, "_server_proc"):
            self._server_proc.terminate()

        if hasattr(self, "_postgresproc"):
            self._postgresproc.stop()

        # ensure children are down
        if hasattr(self, "_server_proc"):
            self._server_proc.wait(20)



"""
    Copyright 2022 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import abc
import json
from abc import ABC
from collections.abc import Sequence
from datetime import datetime
from typing import Generic, Mapping, Optional, TypeVar, Union, cast
from urllib import parse
from urllib.parse import quote
from uuid import UUID

from asyncpg import Record

import inmanta.data
from inmanta import data
from inmanta.data import (
    APILIMIT,
    PRIMITIVE_SQL_TYPES,
    Agent,
    AgentOrder,
    CompileReportOrder,
    ConfigurationModel,
    DatabaseOrderV2,
    DesiredStateVersionOrder,
    DiscoveredResourceOrder,
    FactOrder,
    InvalidQueryParameter,
    InvalidSort,
    Notification,
    NotificationOrder,
    PagingOrder,
    Parameter,
    ParameterOrder,
    QueryFilter,
    Resource,
    ResourceAction,
    ResourceHistoryOrder,
    ResourceLogOrder,
    ResourceStatusOrder,
    SimpleQueryBuilder,
    VersionedResourceOrder,
    model,
)
from inmanta.data.model import (
    BaseModel,
    CompileReport,
    DesiredStateLabel,
    DesiredStateVersion,
    Fact,
    LatestReleasedResource,
    PagingBoundaries,
    ResourceHistory,
    ResourceIdStr,
    ResourceLog,
    ResourceVersionIdStr,
)
from inmanta.protocol.exceptions import BadRequest
from inmanta.protocol.return_value_meta import ReturnValueWithMeta
from inmanta.resources import Id
from inmanta.server import config as opt
from inmanta.server.validate_filter import (
    BooleanEqualityFilter,
    BooleanIsNotNullFilter,
    CombinedContainsFilterResourceState,
    ContainsFilter,
    ContainsFilterResourceAction,
    ContainsPartialFilter,
    DateRangeFilter,
    Filter,
    FilterValidator,
    IntRangeFilter,
    InvalidFilter,
    LogLevelFilter,
)
from inmanta.types import JsonType, SimpleTypes
from inmanta.util import datetime_iso_format

T_ORDER = TypeVar("T_ORDER", bound=DatabaseOrderV2)
T_DTO = TypeVar("T_DTO", bound=BaseModel)


class RequestedPagingBoundaries:
    """Represents the lower and upper bounds that the user requested for the paging boundaries, if any."""

    def __init__(
        self,
        start: Optional[PRIMITIVE_SQL_TYPES],
        end: Optional[PRIMITIVE_SQL_TYPES],
        first_id: Optional[PRIMITIVE_SQL_TYPES],
        last_id: Optional[PRIMITIVE_SQL_TYPES],
    ) -> None:
        self.start = start
        self.end = end
        self.first_id = first_id
        self.last_id = last_id

    def validate(
        self,
    ) -> None:
        start = self.start
        end = self.end
        first_id = self.first_id
        last_id = self.last_id
        if (start is not None) and (end is not None):
            raise InvalidQueryParameter(
                f"Only one of start and end parameters is allowed at the same time. Received start: {start}, end: {end}"
            )
        if (first_id is not None) and (last_id is not None):
            raise InvalidQueryParameter(
                f"Only one of first_id and last_id parameters is allowed at the same time. "
                f"Received first_id: {first_id}, last_id: {last_id}"
            )

        if self.has_start() and self.has_end():
            raise InvalidQueryParameter(
                f"Start and end parameters can not be set at the same time: "
                f"Received first_id: {first_id}, last_id: {last_id}, start: {start}, end: {end}"
            )

    def has_start(self) -> bool:
        return (self.start is not None) or (self.first_id is not None)

    def has_end(self) -> bool:
        return (self.end is not None) or (self.last_id is not None)


class PagingMetadata:
    def __init__(self, total: int, before: int, after: int, page_size: int) -> None:
        self.total = total
        self.before = before
        self.after = after
        self.page_size = page_size

    def to_dict(self) -> dict[str, int]:
        return {
            "total": self.total,
            "before": self.before,
            "after": self.after,
            "page_size": self.page_size,
        }


class DataView(FilterValidator, Generic[T_ORDER, T_DTO], ABC):
    def __init__(
        self,
        order: T_ORDER,
        limit: Optional[int] = None,
        first_id: Optional[PRIMITIVE_SQL_TYPES] = None,
        last_id: Optional[PRIMITIVE_SQL_TYPES] = None,
        start: Optional[PRIMITIVE_SQL_TYPES] = None,
        end: Optional[PRIMITIVE_SQL_TYPES] = None,
        filter: Optional[Mapping[str, Sequence[str]]] = None,
    ) -> None:
        """
        All boundary values for paging are exclusive, i.e. the returned page will contain only values strictly larger/smaller
        than the boundary value. "start", "first", "end" and "last" are poorly named: they imply to be ordering-aware, while
        they're actually only about the requested page containing "larger" or "smaller" values respectively, regardless of
        the requested order.

        :param first_id: secondary boundary value for min bound. If not None, boundary is `> (start, first_id)`.
        :param last_id: secondary boundary value for max bound. If not None, boundary is `< (end, last_id)`.
        :param start: primary boundary value for min bound. If first_id is None, boundary is `> start`.
        :param end: primary boundary value for max bound. If last_id is None, boundary is `< end`.
        """
        self.limit = self.validate_limit(limit)
        self.raw_filter = filter or {}
        self.filter: dict[str, QueryFilter] = self.process_filters(filter)
        self.order = order
        self.requested_page_boundaries = RequestedPagingBoundaries(start, end, first_id, last_id)
        self.requested_page_boundaries.validate()

    @abc.abstractmethod
    def get_base_url(self) -> str:
        """
        Return the base URL used to construct the paging links

        e.g. "/api/v2/resource"
        """

    def get_extra_url_parameters(self) -> dict[str, str]:
        """
        Return additional URL query parameters required to construct the paging links
        """
        return {}

    async def get_data(self) -> tuple[Sequence[T_DTO], Optional[PagingBoundaries]]:
        query_builder = self.get_base_query()

        # In this method, we use `data.Resource`
        # we need the generic functionality of `data.BaseDocument`
        # But that one doesn't actually hold a connection, as it is abstract

        # Project
        query_builder = query_builder.filter(
            *data.Resource.get_composed_filter_with_query_types(
                offset=query_builder.offset, col_name_prefix=None, **self.filter
            )
        )
        query_builder = self.clip_to_page(query_builder)
        sql_query, values = query_builder.build()

        records = await data.Resource.select_query(sql_query, values, no_obj=True)
        dtos = self.construct_dtos(records)

        paging_boundaries = None
        if dtos:
            paging_boundaries = self.order.get_paging_boundaries(dict(records[0]), dict(records[-1]))
        return dtos, paging_boundaries

    @abc.abstractmethod
    def construct_dtos(self, records: Sequence[Record]) -> Sequence[T_DTO]:
        """
        Convert the sequence of records into a sequence of DTO's
        """

    @abc.abstractmethod
    def get_base_query(self) -> SimpleQueryBuilder:
        """
        Return the base query to get the data.

        Must contain select, from and where clause if specific filtering is required
        """

    def get_base_query_for_page_count(self) -> SimpleQueryBuilder:
        """
        Override this method to use a different query than returned by get_base_query()
        to calculate the page count for a certain page.
        """
        return self.get_base_query()

    @property
    @abc.abstractmethod
    def allowed_filters(self) -> dict[str, type[Filter]]:
        """
        Return the specification of the allowed filters, see FilterValidator
        """

    def clip_to_page(self, query_builder: SimpleQueryBuilder) -> SimpleQueryBuilder:
        """
        Update the query builder to constrain it to the page boundaries, order and size
        """
        order = self.order.get_order()
        backward_paging: bool = (order == PagingOrder.ASC and self.requested_page_boundaries.has_end()) or (
            order == PagingOrder.DESC and self.requested_page_boundaries.has_start()
        )

        return (
            query_builder.filter(
                *self.order.as_filter(
                    query_builder.offset,
                    self.requested_page_boundaries.start,
                    self.requested_page_boundaries.first_id,
                    start=True,
                )
            )
            .filter(
                *self.order.as_filter(
                    query_builder.offset,
                    self.requested_page_boundaries.end,
                    self.requested_page_boundaries.last_id,
                    start=False,
                )
            )
            .order_and_limit(self.order, self.limit, backward_paging)
        )

    async def execute(self) -> ReturnValueWithMeta[Sequence[T_DTO]]:
        """
        Main entry point:
        - get the data from the DB
        - get page counts
        - format paging links

        :return: Complete API ReturnValueWithMeta ready to go out
        """
        try:
            dtos, paging_boundaries_in = await self.get_data()

            paging_boundaries: Union[PagingBoundaries, RequestedPagingBoundaries]
            if paging_boundaries_in:
                paging_boundaries = paging_boundaries_in
            else:
                # nothing found now, use the current page boundaries to determine if something exists before us
                paging_boundaries = self.requested_page_boundaries

            metadata = await self._get_page_count(paging_boundaries)
            links = await self.prepare_paging_links(dtos, paging_boundaries, metadata)
            return ReturnValueWithMeta(response=dtos, links=links if links else {}, metadata=metadata.to_dict())
        except (InvalidFilter, InvalidSort, data.InvalidQueryParameter, data.InvalidFieldNameException) as e:
            raise BadRequest(e.message) from e

    # Paging helpers

    async def _get_page_count(self, bounds: Union[PagingBoundaries, RequestedPagingBoundaries]) -> PagingMetadata:
        """
        Construct the page counts,

        either from the PagingBoundaries if we have a valid page,
        or from the RequestedPagingBoundaries if we got an empty page
        """
        query_builder = self.get_base_query_for_page_count()

        query_builder = query_builder.filter(
            *data.BaseDocument.get_composed_filter_with_query_types(
                offset=query_builder.offset, col_name_prefix=None, **self.filter
            )
        )

        order = self.order.get_order()
        reversed = order == PagingOrder.DESC

        end, start, last_id, first_id = bounds.end, bounds.start, bounds.last_id, bounds.first_id

        if reversed:
            end, start = start, end
            last_id, first_id = first_id, last_id

        values = query_builder.values
        after_filter_statements, after_values = self.order.as_filter(len(values) + 1, start, first_id, start=not reversed)
        values.extend(after_values)
        after_filter = data.BaseDocument._join_filter_statements(after_filter_statements)

        before_filter_statements, before_values = self.order.as_filter(len(values) + 1, end, last_id, start=reversed)
        values.extend(before_values)
        before_filter = data.BaseDocument._join_filter_statements(before_filter_statements)

        # If the currently requested page was empty,
        # we are working from RequestedPagingBoundaries instead of PagingBoundaries
        # If the RequestedPagingBoundaries has only nulls, we are not paging and the total count is 0
        # If the RequestedPagingBoundaries has one pair of nulls,
        #   we are paging but the current page is empty and so are the next ones
        # the `as_filter` method will return empty if the input is null
        # so we can conclude that the value for an empty filter is always 0

        if not before_filter and not after_filter:
            # Table is empty
            # Don't even make the query
            return PagingMetadata(
                total=0,
                before=0,
                after=0,
                page_size=self.limit,
            )

        select_clause = (
            "SELECT COUNT(*) as count_total"
            + (f", COUNT(*) filter ({before_filter}) as count_before" if before_filter else "")
            + (f", COUNT(*) filter ({after_filter}) as count_after " if after_filter else "")
        )

        query_builder = query_builder.select(select_clause)

        sql_query, values = query_builder.build()
        result = await data.Resource.select_query(sql_query, values, no_obj=True)
        result = cast(list[Record], result)
        if not result:
            raise InvalidQueryParameter("Could not determine page bounds")
        return PagingMetadata(
            total=cast(int, result[0]["count_total"]),
            before=cast(int, result[0].get("count_before", 0)),
            after=cast(int, result[0].get("count_after", 0)),
            page_size=self.limit,
        )

    async def prepare_paging_links(
        self,
        dtos: Sequence[T_DTO],
        paging_boundaries: Union[PagingBoundaries, RequestedPagingBoundaries],
        meta: PagingMetadata,
    ) -> dict[str, str]:
        """
        Construct the paging links
        """
        links = {}

        url_query_params: dict[str, Optional[Union[SimpleTypes, Sequence[str]]]] = {
            "limit": self.limit,
            "sort": str(self.order),
        }

        for key, value in self.raw_filter.items():
            if value is not None:
                url_query_params[f"filter.{key}"] = value

        url_query_params.update(self.get_extra_url_parameters())

        if dtos:
            base_url = self.get_base_url()

            def value_to_string(value: Union[str, int, UUID, datetime]) -> str:
                if isinstance(value, datetime):
                    # Accross API boundaries, all naive datetime instances are assumed UTC.
                    # Returns ISO timestamp.
                    return datetime_iso_format(value, tz_aware=opt.server_tz_aware_timestamps.get())
                return str(value)

            def make_link(**args: Optional[Union[str, int, UUID, datetime]]) -> str:
                params = url_query_params.copy()
                params.update({k: value_to_string(v) for k, v in args.items() if v is not None})
                return f"{base_url}?{parse.urlencode(params, doseq=True)}"

            link_with_end = make_link(
                end=paging_boundaries.end,
                last_id=paging_boundaries.last_id,
            )
            link_with_start = make_link(
                start=paging_boundaries.start,
                first_id=paging_boundaries.first_id,
            )

            has_next = meta.after > 0
            if has_next:
                if self.order.get_order() == "DESC":
                    links["next"] = link_with_end
                else:
                    links["next"] = link_with_start

            has_prev = meta.before > 0
            if has_prev:
                if self.order.get_order() == "DESC":
                    links["prev"] = link_with_start
                else:
                    links["prev"] = link_with_end
                # First page
                links["first"] = make_link()

            # Same page
            links["self"] = make_link(
                first_id=self.requested_page_boundaries.first_id,
                start=self.requested_page_boundaries.start,
            )
            # TODO: last links
        return links

    def validate_limit(self, limit: Optional[int]) -> int:
        if limit is None:
            return APILIMIT
        if limit > APILIMIT:
            raise BadRequest(f"limit parameter can not exceed {APILIMIT}, got {limit}.")
        return limit


class ResourceView(DataView[ResourceStatusOrder, model.LatestReleasedResource]):
    def __init__(
        self,
        env: data.Environment,
        limit: Optional[int] = None,
        first_id: Optional[ResourceVersionIdStr] = None,
        last_id: Optional[ResourceVersionIdStr] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        filter: Optional[dict[str, list[str]]] = None,
        sort: str = "resource_type.desc",
        deploy_summary: bool = False,
    ) -> None:
        super().__init__(
            order=ResourceStatusOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = env
        self.deploy_summary = deploy_summary

        # Rewrite the filter to special case orphan handling
        # We handle the non-orphan case by changing the query, so we don't need the filter
        # This doesn't affect the paging links, as they use the raw filter

        status_filter_type, status_filter_fields = self.filter.get("status", (None, {}))
        assert status_filter_type is None or status_filter_type == inmanta.data.QueryType.COMBINED
        assert isinstance(status_filter_fields, dict)
        self.drop_orphans = "orphaned" in status_filter_fields.get(inmanta.data.QueryType.NOT_CONTAINS, []) or (
            "orphaned" not in status_filter_fields.get(inmanta.data.QueryType.CONTAINS, ["orphaned"])
        )

        if self.drop_orphans:
            # clean filter for orphans
            try:
                status_filter_fields.get(inmanta.data.QueryType.NOT_CONTAINS, []).remove("orphaned")
                if not status_filter_fields.get(inmanta.data.QueryType.NOT_CONTAINS, []):
                    del status_filter_fields[inmanta.data.QueryType.NOT_CONTAINS]
                if not status_filter_fields:
                    del self.filter["status"]
            except ValueError:
                pass

    @property
    def allowed_filters(self) -> dict[str, type[Filter]]:
        return {
            "resource_type": ContainsPartialFilter,
            "agent": ContainsPartialFilter,
            "resource_id_value": ContainsPartialFilter,
            "status": CombinedContainsFilterResourceState,
        }

    def get_base_url(self) -> str:
        return "/api/v2/resource"

    def get_extra_url_parameters(self) -> dict[str, str]:
        return {"deploy_summary": str(self.deploy_summary)}

    def get_base_query(self) -> SimpleQueryBuilder:
        new_query_builder = SimpleQueryBuilder(
            select_clause="SELECT *",
            prelude=f"""
               WITH latest_version AS (
                    SELECT MAX(public.configurationmodel.version) as version
                    FROM public.configurationmodel
                    WHERE public.configurationmodel.released=TRUE AND environment=$1
                ), versioned_resource_state AS (
                    SELECT
                        rps.*,
                        CASE
                            -- try the cheap, trivial option first because the lookup has a big performance impact
                            WHEN EXISTS (
                                SELECT 1
                                FROM resource AS r
                                WHERE r.environment = rps.environment AND r.resource_id = rps.resource_id AND r.model = (
                                    SELECT version FROM latest_version
                                )
                            ) THEN (SELECT version FROM latest_version)
                            -- only if the resource does not exist in the latest released version, search for the latest
                            -- version it does exist in
                            ELSE (
                                SELECT MAX(r.model)
                                FROM resource AS r
                                JOIN configurationmodel AS m
                                    ON r.environment = m.environment AND r.model = m.version AND m.released = TRUE
                                WHERE r.environment = rps.environment AND r.resource_id = rps.resource_id
                            )
                        END AS version
                    FROM resource_persistent_state AS rps
                    WHERE rps.environment = $1
                ), result AS (
                    SELECT
                        rps.resource_id,
                        r.attributes,
                        rps.resource_type,
                        rps.agent,
                        rps.resource_id_value,
                        r.model,
                        rps.environment,
                        (
                            CASE
                                -- The resource_persistent_state.last_non_deploying_status column is only populated for
                                -- actual deployment operations to prevent locking issues. This case-statement calculates
                                -- the correct state from the combination of the resource table and the
                                -- resource_persistent_state table.
                                WHEN r.model < (SELECT version FROM latest_version)
                                    THEN 'orphaned'
                                WHEN r.status::text IN('deploying', 'undefined', 'skipped_for_undefined')
                                    -- The deploying, undefined and skipped_for_undefined states are not tracked in the
                                    -- resource_persistent_state table.
                                    THEN r.status::text
                                WHEN rps.last_deployed_attribute_hash != r.attribute_hash
                                    -- The hash changed since the last deploy -> new desired state
                                    THEN r.status::text
                                    -- No override required, use last known state from actual deployment
                                    ELSE rps.last_non_deploying_status::text
                            END
                        ) as status
                    FROM versioned_resource_state AS rps
            -- LEFT join for trivial `COUNT(*)`. Not applicable when filtering orphans because left table contains orphans.
                    {'' if self.drop_orphans else 'LEFT'} JOIN resource AS r
                        ON r.environment = rps.environment
                          AND r.resource_id = rps.resource_id
           -- shortcut the version selection to the latest one iff we wish to exclude orphans
           -- => no per-resource MAX required + wider index application
                          AND r.model = {'(SELECT version FROM latest_version)' if self.drop_orphans else 'rps.version'}
                    WHERE rps.environment = $1
                )
            """,
            from_clause="FROM result AS r",
            values=[self.environment.id],
        )
        return new_query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[model.LatestReleasedResource]:
        dtos: Sequence[LatestReleasedResource] = [
            model.LatestReleasedResource(
                resource_id=resource["resource_id"],
                resource_version_id=resource["resource_id"] + ",v=" + str(resource["model"]),
                id_details=data.Resource.get_details_from_resource_id(resource["resource_id"]),
                status=resource["status"],
                requires=json.loads(resource["attributes"]).get("requires", []),
            )
            for resource in records
            if resource["attributes"]  # filter out bad joins
        ]
        return dtos


class ResourcesInVersionView(DataView[VersionedResourceOrder, model.VersionedResource]):
    def __init__(
        self,
        environment: data.Environment,
        version: int,
        limit: Optional[int] = None,
        filter: Optional[dict[str, list[str]]] = None,
        sort: str = "resource_type.desc",
        first_id: Optional[ResourceVersionIdStr] = None,
        last_id: Optional[ResourceVersionIdStr] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
    ) -> None:
        super().__init__(
            order=VersionedResourceOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment
        self.version = version

    # Per view config
    def get_base_url(self) -> str:
        return f"/api/v2/desiredstate/{self.version}"

    @property
    def allowed_filters(self) -> dict[str, type[Filter]]:
        return {
            "resource_type": ContainsPartialFilter,
            "agent": ContainsPartialFilter,
            "resource_id_value": ContainsPartialFilter,
        }

    def get_base_query(self) -> SimpleQueryBuilder:
        query_builder = SimpleQueryBuilder(
            select_clause="SELECT resource_id, attributes, resource_type, agent, resource_id_value, environment",
            from_clause=f" FROM {data.Resource.table_name()}",
            filter_statements=["environment = $1", "model = $2"],
            values=[self.environment.id, self.version],
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[model.VersionedResource]:
        return [
            model.VersionedResource(
                resource_id=versioned_resource["resource_id"],
                resource_version_id=versioned_resource["resource_id"] + f",v={self.version}",
                id_details=data.Resource.get_details_from_resource_id(versioned_resource["resource_id"]),
                requires=json.loads(versioned_resource["attributes"]).get("requires", []),  # todo: broken
            )
            for versioned_resource in records
        ]


class CompileReportView(DataView[CompileReportOrder, CompileReport]):
    def __init__(
        self,
        environment: data.Environment,
        limit: Optional[int] = None,
        filter: Optional[dict[str, list[str]]] = None,
        sort: str = "resource_type.desc",
        first_id: Optional[UUID] = None,
        last_id: Optional[UUID] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> None:
        super().__init__(
            order=CompileReportOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment

    @property
    def allowed_filters(self) -> dict[str, type[Filter]]:
        return {
            "requested": DateRangeFilter,
            "success": BooleanEqualityFilter,
            "started": BooleanIsNotNullFilter,
            "completed": BooleanIsNotNullFilter,
        }

    def get_base_url(self) -> str:
        return "/api/v2/compilereport"

    def get_base_query(self) -> SimpleQueryBuilder:
        query_builder = SimpleQueryBuilder(
            select_clause="""SELECT id, remote_id, environment, requested,
                            started, completed, do_export, force_update,
                            metadata, requested_environment_variables, used_environment_variables,
                            mergeable_environment_variables,
                            success, version, partial, removed_resource_sets, exporter_plugin,
                            notify_failed_compile, failed_compile_message""",
            from_clause=f" FROM {data.Compile.table_name()}",
            filter_statements=["environment = $1"],
            values=[self.environment.id],
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[model.CompileReport]:
        return [
            CompileReport(
                id=compile["id"],
                remote_id=compile["remote_id"],
                environment=compile["environment"],
                requested=compile["requested"],
                started=compile["started"],
                completed=compile["completed"],
                success=compile["success"],
                version=compile["version"],
                do_export=compile["do_export"],
                force_update=compile["force_update"],
                metadata=json.loads(compile["metadata"]) if compile["metadata"] else {},
                environment_variables=(
                    json.loads(compile["used_environment_variables"]) if compile["used_environment_variables"] else None
                ),
                requested_environment_variables=json.loads(compile["requested_environment_variables"]),
                mergeable_environment_variables=json.loads(compile["mergeable_environment_variables"]),
                partial=compile["partial"],
                removed_resource_sets=compile["removed_resource_sets"],
                exporter_plugin=compile["exporter_plugin"],
                notify_failed_compile=compile["notify_failed_compile"],
                failed_compile_message=compile["failed_compile_message"],
            )
            for compile in records
        ]


class DesiredStateVersionView(DataView[DesiredStateVersionOrder, DesiredStateVersion]):
    def __init__(
        self,
        environment: data.Environment,
        limit: Optional[int] = None,
        filter: Optional[dict[str, list[str]]] = None,
        sort: str = "resource_type.desc",
        start: Optional[int] = None,
        end: Optional[int] = None,
    ) -> None:
        super().__init__(
            order=DesiredStateVersionOrder.parse_from_string(sort),
            limit=limit,
            first_id=None,
            last_id=None,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment

    @property
    def allowed_filters(self) -> dict[str, type[Filter]]:
        return {
            "version": IntRangeFilter,
            "date": DateRangeFilter,
            "status": ContainsFilter,
        }

    def get_base_url(self) -> str:
        return "/api/v2/desiredstate"

    def get_base_query(self) -> SimpleQueryBuilder:
        subquery, subquery_values = ConfigurationModel.desired_state_versions_subquery(self.environment.id)
        query_builder = SimpleQueryBuilder(
            select_clause="SELECT *",
            from_clause=f" FROM ({subquery}) as result",
            values=subquery_values,
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[DesiredStateVersion]:
        return [
            DesiredStateVersion(
                version=desired_state["version"],
                date=desired_state["date"],
                total=desired_state["total"],
                labels=(
                    [DesiredStateLabel(name=desired_state["type"], message=desired_state["message"])]
                    if desired_state["type"] and desired_state["message"]
                    else []
                ),
                status=desired_state["status"],
            )
            for desired_state in records
        ]


class ResourceHistoryView(DataView[ResourceHistoryOrder, ResourceHistory]):
    def __init__(
        self,
        environment: data.Environment,
        rid: ResourceIdStr,
        limit: Optional[int] = None,
        sort: str = "resource_type.desc",
        first_id: Optional[str] = None,
        last_id: Optional[str] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> None:
        super().__init__(
            order=ResourceHistoryOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter={},
        )
        self.environment = environment
        self.rid = rid

    @property
    def allowed_filters(self) -> dict[str, type[Filter]]:
        return {}

    def get_base_url(self) -> str:
        return f"/api/v2/resource/{quote(self.rid, safe='')}/history"

    def get_base_query(self) -> SimpleQueryBuilder:
        query_builder = SimpleQueryBuilder(
            prelude="""
            /* Assign a sequence id to the rows, which is used for grouping consecutive ones with the same hash */
                WITH resourcewithsequenceids AS (
                  SELECT
                    attribute_hash,
                    model,
                    attributes,
                    date,
                    ROW_NUMBER() OVER (ORDER BY date) - ROW_NUMBER() OVER (
                      PARTITION BY attribute_hash
                      ORDER BY date
                    ) AS seqid
                  FROM resource JOIN configurationmodel cm
                    ON resource.model = cm.version AND resource.environment = cm.environment
                  WHERE resource.environment = $1 AND resource_id = $2 AND cm.released = TRUE
                )
            """,
            select_clause="SELECT attribute_hash, date, attributes, model",
            from_clause="""
            FROM (SELECT
                    attribute_hash,
                    min(date) as date,
                    min(model) as model,
                        (SELECT distinct on (attribute_hash) attributes
                            FROM resourcewithsequenceids
                            WHERE resourcewithsequenceids.attribute_hash = rs.attribute_hash
                            AND resourcewithsequenceids.seqid = rs.seqid
                            ORDER BY attribute_hash, model
                        ) as attributes
                    FROM resourcewithsequenceids rs
                    GROUP BY attribute_hash,  seqID) as sub
            """,
            values=[self.environment.id, self.rid],
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[ResourceHistory]:
        def get_attributes(record: Record) -> JsonType:
            attributes = json.loads(record["attributes"])
            if "version" not in attributes:
                # Due to a bug, the version field has always been present in the attributes dictionary.
                # This bug has been fixed in the database. For backwards compatibility reason we here make sure that the
                # version field is present in the attributes dictionary served out via the API.
                attributes["version"] = record["model"]
            return attributes

        return [
            ResourceHistory(
                resource_id=self.rid,
                attribute_hash=record["attribute_hash"],
                attributes=get_attributes(record),
                date=record["date"],
                requires=[Id.parse_id(rid).resource_str() for rid in json.loads(record["attributes"]).get("requires", [])],
            )
            for record in records
        ]


class ResourceLogsView(DataView[ResourceLogOrder, ResourceLog]):
    def __init__(
        self,
        environment: data.Environment,
        rid: ResourceIdStr,
        limit: Optional[int] = None,
        sort: str = "resource_type.desc",
        filter: Optional[dict[str, list[str]]] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> None:
        super().__init__(
            order=ResourceLogOrder.parse_from_string(sort),
            limit=limit,
            first_id=None,
            last_id=None,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment
        self.rid = rid

    @property
    def allowed_filters(self) -> dict[str, type[Filter]]:
        return {
            "minimal_log_level": LogLevelFilter,
            "timestamp": DateRangeFilter,
            "message": ContainsPartialFilter,
            "action": ContainsFilterResourceAction,
        }

    def process_filters(self, filter: Optional[Mapping[str, Sequence[str]]]) -> dict[str, QueryFilter]:
        # Change the api names of the filters to the names used internally in the database
        query = super().process_filters(filter)
        if query.get("minimal_log_level"):
            filter_value = query.pop("minimal_log_level")
            query["level"] = filter_value
        if query.get("message"):
            filter_value = query.pop("message")
            query["msg"] = filter_value
        return query

    def get_base_url(self) -> str:
        return f"/api/v2/resource/{parse.quote(self.rid, safe='')}/logs"

    def get_base_query(self) -> SimpleQueryBuilder:
        query_builder = SimpleQueryBuilder(
            prelude=f"""
                -- Get all resource action in the given environment for the given resource_id
                WITH actions AS (
                    SELECT  ra.*
                    FROM {Resource.table_name()} AS r INNER JOIN resourceaction_resource AS rr ON (
                                                          r.environment=rr.environment
                                                          AND r.resource_id=rr.resource_id
                                                          AND r.model=rr.resource_version
                                                      )
                                                      INNER JOIN {ResourceAction.table_name()} AS ra ON (
                                                          rr.resource_action_id=ra.action_id
                                                      )
                    WHERE r.environment=$1 AND r.resource_id=$2
                )
            """,
            select_clause="SELECT action_id, action, timestamp, unnested_message",
            from_clause="""
            FROM
                (
                    SELECT action_id,
                           action,
                           (unnested_message ->> 'timestamp')::timestamptz AS timestamp,
                           unnested_message ->> 'level' AS level,
                           unnested_message ->> 'msg' AS msg,
                           unnested_message
                    FROM actions, unnest(messages) AS unnested_message
                ) AS unnested
            """,
            values=[self.environment.id, self.rid],
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[ResourceLog]:
        logs = []
        for record in records:
            message = json.loads(record["unnested_message"])
            logs.append(
                ResourceLog(
                    action_id=record["action_id"],
                    action=record["action"],
                    timestamp=record["timestamp"],
                    level=message.get("level"),
                    msg=message.get("msg"),
                    args=message.get("args", []),
                    kwargs=message.get("kwargs", {}),
                )
            )
        return logs


class FactsView(DataView[FactOrder, Fact]):
    def __init__(
        self,
        environment: data.Environment,
        limit: Optional[int] = None,
        sort: str = "resource_type.desc",
        first_id: Optional[UUID] = None,
        last_id: Optional[UUID] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        filter: Optional[dict[str, list[str]]] = None,
    ) -> None:
        super().__init__(
            order=FactOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment

    @property
    def allowed_filters(self) -> dict[str, type[Filter]]:
        return {
            "name": ContainsPartialFilter,
            "resource_id": ContainsPartialFilter,
            "expires": BooleanEqualityFilter,
        }

    def get_base_url(self) -> str:
        return "/api/v2/facts"

    def get_base_query(self) -> SimpleQueryBuilder:
        query_builder = SimpleQueryBuilder(
            select_clause="SELECT p.id, p.name, p.value, p.source, p.resource_id, p.updated, p.metadata, p.environment",
            from_clause=f"FROM {Parameter.table_name()} as p",
            filter_statements=["p.environment = $1 ", "p.source = 'fact'"],
            values=[self.environment.id],
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[Fact]:
        return [
            Fact(
                id=fact["id"],
                name=fact["name"],
                value=fact["value"],
                source=fact["source"],
                updated=fact["updated"],
                resource_id=fact["resource_id"],
                metadata=json.loads(fact["metadata"]) if fact["metadata"] else None,
                environment=fact["environment"],
            )
            for fact in records
        ]


class NotificationsView(DataView[NotificationOrder, model.Notification]):
    def __init__(
        self,
        environment: data.Environment,
        limit: Optional[int] = None,
        sort: str = "resource_type.desc",
        first_id: Optional[UUID] = None,
        last_id: Optional[UUID] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        filter: Optional[dict[str, list[str]]] = None,
    ) -> None:
        super().__init__(
            order=NotificationOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment

    @property
    def allowed_filters(self) -> dict[str, type[Filter]]:
        return {
            "title": ContainsPartialFilter,
            "message": ContainsPartialFilter,
            "read": BooleanEqualityFilter,
            "cleared": BooleanEqualityFilter,
            "severity": ContainsFilter,
        }

    def get_base_url(self) -> str:
        return "/api/v2/notification"

    def get_base_query(self) -> SimpleQueryBuilder:
        return SimpleQueryBuilder(
            select_clause="""SELECT n.*""",
            from_clause=f" FROM {Notification.table_name()} as n",
            filter_statements=[" environment = $1 "],
            values=[self.environment.id],
        )

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[model.Notification]:
        return [
            model.Notification(
                id=notification["id"],
                title=notification["title"],
                message=notification["message"],
                severity=notification["severity"],
                created=notification["created"],
                read=notification["read"],
                cleared=notification["cleared"],
                uri=notification["uri"],
                environment=notification["environment"],
            )
            for notification in records
        ]


class ParameterView(DataView[ParameterOrder, model.Parameter]):
    def __init__(
        self,
        environment: data.Environment,
        limit: Optional[int] = None,
        sort: str = "resource_type.desc",
        first_id: Optional[UUID] = None,
        last_id: Optional[UUID] = None,
        start: Optional[Union[str, datetime]] = None,
        end: Optional[Union[str, datetime]] = None,
        filter: Optional[dict[str, list[str]]] = None,
    ) -> None:
        super().__init__(
            order=ParameterOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment

    @property
    def allowed_filters(self) -> dict[str, type[Filter]]:
        return {
            "name": ContainsPartialFilter,
            "source": ContainsPartialFilter,
            "updated": DateRangeFilter,
        }

    def get_base_url(self) -> str:
        return "/api/v2/parameters"

    def get_base_query(self) -> SimpleQueryBuilder:
        return SimpleQueryBuilder(
            select_clause="""SELECT p.id, p.name, p.value, p.source, p.updated, p.metadata, p.environment""",
            from_clause=f"FROM {Parameter.table_name()} as p",
            filter_statements=["environment = $1", "p.source != 'fact'"],
            values=[self.environment.id],
        )

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[model.Parameter]:
        return [
            model.Parameter(
                id=parameter["id"],
                name=parameter["name"],
                value=parameter["value"],
                source=parameter["source"],
                updated=parameter["updated"],
                metadata=json.loads(parameter["metadata"]) if parameter["metadata"] else None,
                environment=parameter["environment"],
            )
            for parameter in records
        ]


class AgentView(DataView[AgentOrder, model.Agent]):
    def __init__(
        self,
        environment: data.Environment,
        limit: Optional[int] = None,
        sort: str = "resource_type.desc",
        start: Optional[Union[datetime, bool, str]] = None,
        end: Optional[Union[datetime, bool, str]] = None,
        first_id: Optional[str] = None,
        last_id: Optional[str] = None,
        filter: Optional[dict[str, list[str]]] = None,
    ) -> None:
        super().__init__(
            order=AgentOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment

    @property
    def allowed_filters(self) -> dict[str, type[Filter]]:
        return {
            "name": ContainsPartialFilter,
            "process_name": ContainsPartialFilter,
            "status": ContainsFilter,
        }

    def process_filters(self, filter: Optional[Mapping[str, Sequence[str]]]) -> dict[str, QueryFilter]:
        out_filter = super().process_filters(filter)
        # name is ambiguous, qualify
        if "name" in out_filter:
            out_filter["a.name"] = out_filter.pop("name")
        return out_filter

    def get_base_url(self) -> str:
        return "/api/v2/agents"

    def get_base_query(self) -> SimpleQueryBuilder:
        base = SimpleQueryBuilder(
            select_clause="""SELECT a.name, a.environment, last_failover, paused, unpause_on_resume,
                                            ap.hostname as process_name, ai.process as process_id,
                                            (CASE WHEN paused THEN 'paused'
                                                WHEN id_primary IS NOT NULL THEN 'up'
                                                ELSE 'down'
                                            END) as status""",
            from_clause=f" FROM {Agent.table_name()} as a LEFT JOIN public.agentinstance ai ON a.id_primary=ai.id "
            " LEFT JOIN public.agentprocess ap ON ai.process = ap.sid",
            filter_statements=[" a.environment = $1 "],
            values=[self.environment.id],
        )
        # wrap when using compound fields
        virtual_fields = {"status", "process_name", "process_id"}
        used_fields = set(self.filter.keys()).union({t[0] for t in self.order.get_order_elements(False)})
        if virtual_fields.intersection(used_fields):
            query, values = base.build()
            return SimpleQueryBuilder(
                select_clause="select *",
                from_clause=f"FROM ({query}) as a",
                values=values,
            )
        return base

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[model.Agent]:
        return [
            model.Agent(
                name=agent["name"],
                environment=agent["environment"],
                last_failover=agent["last_failover"],
                paused=agent["paused"],
                unpause_on_resume=agent["unpause_on_resume"],
                process_id=agent["process_id"],
                process_name=agent["process_name"],
                status=agent["status"],
            )
            for agent in records
        ]


class DiscoveredResourceView(DataView[DiscoveredResourceOrder, model.DiscoveredResource]):
    def __init__(
        self,
        environment: data.Environment,
        limit: Optional[int] = None,
        sort: str = "discovered_resource_id.asc",
        start: Optional[str] = None,
        end: Optional[str] = None,
        filter: Optional[Mapping[str, Sequence[str]]] = None,
    ) -> None:
        super().__init__(
            order=DiscoveredResourceOrder.parse_from_string(sort),
            limit=limit,
            first_id=None,
            last_id=None,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment

    @property
    def allowed_filters(self) -> dict[str, type[Filter]]:
        """
        Return the specification of the allowed filters, see FilterValidator
        """
        return {
            "managed": BooleanEqualityFilter,
        }

    def get_base_url(self) -> str:
        return "/api/v2/discovered"

    def get_base_query(self) -> SimpleQueryBuilder:
        query_builder = SimpleQueryBuilder(
            select_clause="SELECT *",
            prelude=f"""
            WITH result AS (
                SELECT
                    dr.environment,
                    dr.discovered_resource_id,
                    dr.values,
                    (rps.resource_id IS NOT NULL)  AS managed

                FROM {data.DiscoveredResource.table_name()} as dr
                LEFT JOIN {data.ResourcePersistentState.table_name()} rps
                ON dr.environment=rps.environment AND dr.discovered_resource_id = rps.resource_id
                WHERE dr.environment = $1
            )
            """,
            from_clause="FROM result AS r",
            values=[self.environment.id],
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[dict[str, str]]:
        return [
            model.DiscoveredResource(
                discovered_resource_id=res["discovered_resource_id"],
                values=json.loads(res["values"]),
                managed_resource_uri=f"/api/v2/resource/{res['discovered_resource_id']}" if res["managed"] else None,
            ).model_dump()
            for res in records
        ]


class PreludeBasedFilteringQueryBuilder(SimpleQueryBuilder):
    """
    A query builder that applies any filters and the LIMIT statement to the prelude query rather than the outer query.
    The outer query may use the table name "prelude" to refer to the inner query.
    """

    def __init__(
        self,
        prelude_query_builder: SimpleQueryBuilder,
        select_clause: Optional[str] = None,
        from_clause: Optional[str] = None,
        db_order: Optional[DatabaseOrderV2] = None,
        backward_paging: bool = False,
    ) -> None:
        super().__init__(
            select_clause=select_clause,
            from_clause=from_clause,
            filter_statements=None,
            values=None,
            db_order=db_order,
            limit=None,
            backward_paging=backward_paging,
            prelude=None,
        )
        self._prelude_query_builder = prelude_query_builder

    @property
    def offset(self) -> int:
        """The current offset of the values to be used for filter statements"""
        return len(self.values) + len(self._prelude_query_builder.values) + 1

    def build(self) -> tuple[str, list[object]]:
        prelude_query, prelude_values = self._prelude_query_builder.build()
        prelude_query_in_with_block = f"WITH prelude AS ({prelude_query})"
        delegate: SimpleQueryBuilder = SimpleQueryBuilder(
            select_clause=self.select_clause,
            from_clause=self._from_clause,
            filter_statements=self._prelude_query_builder.filter_statements + self.filter_statements,
            values=self._prelude_query_builder.values + self.values,
            db_order=self.db_order,
            limit=None,
            backward_paging=self.backward_paging,
            prelude=prelude_query_in_with_block,
        )
        full_query, values_full = delegate.build()
        return full_query, prelude_values

    def select(self, select_clause: str) -> "PreludeBasedFilteringQueryBuilder":
        """Set the select clause of the query"""
        return PreludeBasedFilteringQueryBuilder(
            select_clause=select_clause,
            from_clause=self._from_clause,
            db_order=self.db_order,
            backward_paging=self.backward_paging,
            prelude_query_builder=self._prelude_query_builder,
        )

    def from_clause(self, from_clause: str) -> "PreludeBasedFilteringQueryBuilder":
        """Set the from clause of the query"""
        return PreludeBasedFilteringQueryBuilder(
            select_clause=self.select_clause,
            from_clause=from_clause,
            db_order=self.db_order,
            backward_paging=self.backward_paging,
            prelude_query_builder=self._prelude_query_builder,
        )

    def order_and_limit(
        self, db_order: DatabaseOrderV2, limit: Optional[int] = None, backward_paging: bool = False
    ) -> "PreludeBasedFilteringQueryBuilder":
        """Set the order and limit of the query"""
        return PreludeBasedFilteringQueryBuilder(
            select_clause=self.select_clause,
            from_clause=self._from_clause,
            db_order=db_order,
            backward_paging=backward_paging,
            prelude_query_builder=self._prelude_query_builder.order_and_limit(db_order, limit, backward_paging),
        )

    def filter(self, filter_statements: list[str], values: list[object]) -> "PreludeBasedFilteringQueryBuilder":
        return PreludeBasedFilteringQueryBuilder(
            select_clause=self.select_clause,
            from_clause=self._from_clause,
            db_order=self.db_order,
            backward_paging=self.backward_paging,
            prelude_query_builder=self._prelude_query_builder.filter(filter_statements, values),
        )
